.TH sql.ppg Onyx
.SH NAME
sql.ppg \- Base grammar for SQL's
.SH SYNOPSYIS
.nf
class Sql : public Parser {
	Sql(Callback *);
	~Sql();
	
	void operator	<<(char *);
	int 		commit();
	
	private :
	int transaction();
	int statement();
	
	int trade();
	int create();
	int drop();
	int grant();
	int revoke();
	int insert();
	int select();
	int update();
	int del();
	int load();
	int unload();
	
	int tabid();
	int colid();
	int indid();
	int svalue();
	int value();
	int compop();
	int boolop();
	int boolop();
	int numop();
	int expression();
	int condition();
	}
.fi
.SH IMPLEMENTATION
.PP
transaction : 
.br
.in +0.8i
.ad l
.nh
( statement  ";" ) 
.br
;
.br
.hy
.PP
statement : 
.br
.in +0.8i
.ad l
.nh
trade  | create  | drop  | grant  | revoke  | insert  | select  | update  | del  | load  | unload  
.br
;
.br
.hy
.PP
trade : 
.br
.in +0.8i
.ad l
.nh
"user" id  [ "@" matchword  ] [ "ticket" matchword  ] | "connect" id  [ "%" id  ] [ "@" matchword  ] | "disconnect" | "begin" matchword  | "commit" matchword  
.br
;
.br
.hy
.PP
create : 
.br
.in +0.8i
.ad l
.nh
"create" [ "table" id  "(" ( "constraint" id  "check" "(" match  ")" | "primary" "key" "(" ( id  ! "," ) ")" | "foreign" "key" "(" ( id  ! "," ) ")" "references" id  [ "on" [ "delete" | "update" | "true" ] [ "cascade" | "restricted" | "set" svalue  ] ] | id  type  [ "not" "null" ] ! "," ) ")" matchexp  | "view" id  "as" select  | "index" "on" tabid  "(" ( colid  ! "," ) ")" | true  error  ] 
.br
;
.br
.hy
.PP
drop : 
.br
.in +0.8i
.ad l
.nh
"drop" [ "table" tabid  | "view" tabid  | "index" indid  | true  error  ] 
.br
;
.br
.hy
.PP
grant : 
.br
.in +0.8i
.ad l
.nh
"grant" id  "on" tabid  "to" id  
.br
;
.br
.hy
.PP
revoke : 
.br
.in +0.8i
.ad l
.nh
"revoke" id  "on" tabid  "from" id  
.br
;
.br
.hy
.PP
insert : 
.br
.in +0.8i
.ad l
.nh
"insert" "into" tabid  [ "(" ( colid  ! "," ) ")" ] "values" "(" ( svalue  ! "," ) ")" 
.br
;
.br
.hy
.PP
select : 
.br
.in +0.8i
.ad l
.nh
"select" ( colid  | "*" ! "," ) "from" ( tabid  ! "," ) "where" condition  
.br
;
.br
.hy
.PP
delete : 
.br
.in +0.8i
.ad l
.nh
"delete" "from" tabid  "where" condition  
.br
;
.br
.hy
.PP
update : 
.br
.in +0.8i
.ad l
.nh
"update" tabid  "set" ( colid  "=" expression  ! "," ) "where" condition  
.br
;
.br
.hy
.PP
load : 
.br
.in +0.8i
.ad l
.nh
"load" "from" "\(rs"" match  "\(rs"" "insert" "into" tabid  [ "commit" num  ] 
.br
;
.br
.hy
.PP
unload : 
.br
.in +0.8i
.ad l
.nh
"unload" "to" "\(rs"" match  "\(rs"" select  
.br
;
.br
.hy
.PP
tabid : 
.br
.in +0.8i
.ad l
.nh
id  
.br
;
.br
.hy
.PP
colid : 
.br
.in +0.8i
.ad l
.nh
id  
.br
;
.br
.hy
.PP
indid : 
.br
.in +0.8i
.ad l
.nh
id  
.br
;
.br
.hy
.PP
svalue : 
.br
.in +0.8i
.ad l
.nh
num  | "\(rs"" match  "\(rs"" | "'" match  "'" 
.br
;
.br
.hy
.PP
value : 
.br
.in +0.8i
.ad l
.nh
svalue  | "not" expression  expression  | "(" expression  ")" | colid  
.br
;
.br
.hy

A value can be a string, a number, a not expression, a
expression in braces, or a identifier.
.PP
operator : 
.br
.in +0.8i
.ad l
.nh
"=" | "!=" | "<>" | "<=" | ">=" | "<" | ">" | "like" 
.br
;
.br
.hy

Here the different compare opperators are translated.
Eg. "like" is translated to "~", and "=" is translated
to "==".
.PP
boolop : 
.br
.in +0.8i
.ad l
.nh
"and" | "or" 
.br
;
.br
.hy

Here the boolean operators are translated.
.PP
numop : 
.br
.in +0.8i
.ad l
.nh
"+" | "-" | "*" | "/" | "%" | "mod" 
.br
;
.br
.hy

Here the numeric operators are translated.
.PP
expression : 
.br
.in +0.8i
.ad l
.nh
value  ( compop  value  | boolop  value  | numop  value  ) 
.br
;
.br
.hy

A expression starts with a value and perhaps continues with
operator value pairs.
.PP
condition : 
.br
.in +0.8i
.ad l
.nh
expression  
.br
;
.br
.hy

A condition is nothing else than a expression. This is done
because PPG is only LL(1) and I dont have any better solution
in mind.

Caution no type checking is done, perhaps in a later version.
