/* ----------------------------------------------------------------------------
 * scrolls.C (c) Michael Krahe
 */

extern "C" {
#include <stdio.h>
#include <string.h>
#include <memory.h>
#ifdef _BSDI
#include <stdlib.h>
#else
#include <malloc.h>
#endif
#include <signal.h>
}
#include "twin.h"

ViewScroll::ViewScroll(char *InitName,View *Panner,
		int InitX, int InitY, int InitB,int InitH)
        : View(InitName,InitX,InitY,InitB,InitH)
{	panner=Panner;
	xo = 0; yo = 0;
	border = 1;
	footer = 0;
	header = 0;

	freepos=!InitX || !InitY;

	if (panner) {
		panner->x = 0;
		panner->y = 0;
		panner->parent = this;
		if (!panner->b) panner->b = InitB-2;
		if (!panner->h) panner->h = InitH-2;
		if (!InitB) b = panner->b +2;
		if (!InitH) h = panner->h +2;
		}
	}

void ViewScroll::move(int MoveX,int MoveY)
{       int v;

        v = visible;
        if (v) {
                unmap();
                VRoot.expose(View::RealX(0),View::RealY(0),b,h);
                }
        x=MoveX;
        y=MoveY;
        if (v) {
                map(parent);
                VRoot.expose(View::RealX(0),View::RealY(0),b,h);
                }
        }

void ViewScroll::map(View *par)
{	ViewList *vl;
	int nx;

	if (panner) panner->visible=1;

	if (!x && !y && border) {
		if (!b) b = panner->b +2;
		if (!h) h = panner->h +2;

		nx=3;
		do {	vl = VRoot.viewlist;
			while (vl) {
				if (vl->view->x == nx) break;
				vl = vl->next;
				}
			if (!vl) break;
			nx++;
			} while (nx<max_y-4);
		if (nx>=max_y-4) nx=3;
		x = nx;
		y = nx;

		if (x+b > max_x-3) b=(max_x-x)-2;
		if (y+h > max_y-3) h=(max_y-y)-2;
		}
	if (!x && !y && !border) {
		if (!b) b = panner->b +2;
		if (!h) h = panner->h +2;

		x=1;
		y=2+header;
		b=max_x-2;
		h=max_y-(2+footer+header);
		}

        if (par) parent =par;
        if (!visible) {
                visible = 1;
                VRoot.expose(View::RealX(0),View::RealY(0),b,h);
                }
        else    visible = 1;
	}

void ViewScroll::unmap()
{	if (panner) panner->visible=0;

        if (visible) {
                visible = 0;
                VRoot.expose(View::RealX(0),View::RealY(0),b,h);
                }
        else    visible = 0;
	if (freepos) { x=0; y=0; }
	}

int ViewScroll::RealX(int vx)
{       if (parent)
                return(parent->RealX(vx+x)+1-xo);
        else    return(vx+x);
        }

int ViewScroll::RealY(int vy)
{       if (parent)
                return(parent->RealY(vy+y)+1-yo);
        else    return(vy+y);
        }

void ViewScroll::set_cursor(int cx,int cy)
{       int exp_flag=0;

	if (visible && parent) {
		if (cx-xo < 0) {
			exp_flag=xo-cx;
			xo-=exp_flag;
			}
		if (cy-yo < 0) {
			exp_flag=yo-cy;
			yo-=exp_flag;
			}
		if (cx-xo >= b-2) {
			exp_flag=(cx-xo)-(b-3);
			xo+=exp_flag;
			}
		if (cy-yo >= h-2) {
			exp_flag=(cy-yo)-(h-3);
			yo+=exp_flag;
			}
		if (exp_flag) VRoot.expose(View::RealX(0),View::RealY(0),b,h);
                parent->set_cursor(cx+x+1-xo,cy+y+1-yo);
		}
        }

void ViewScroll::expose(int ExX,int ExY,int ExB,int ExH)
{       int ex,ey,eb,eh;
	int i;
	char *line;
	
	if (!visible) return;

        if (!ExB || (ExB+ExX > b)) ExB = b-ExX;
        if (!ExH || (ExH+ExY > h)) ExH = h-ExY;

        line = (char*)setmalloc(b+1);

	if (!border)
       	for (i=0; i<h; i++) {
                if (footer && (i == h-1))
                        memset(line,'q'+0x20,b);
                else
                if (header && (i == 0))
                        memset(line,'q'+0x20,b);
                else	memset(line,' ',b);

                line[b]=0;

                if ((i >= ExY) && (i < ExY+ExH)) {
                        line[ExX+ExB] = 0;
                        tw_term_puts(View::RealX(ExX),View::RealY(i),TW_NORMAL,line+ExX);
                        }
                if (i >= ExY+ExH) break;
        	}
	else
        for (i=0; i<h; i++) {
                if ((i == 0) || (i == h-1)) {
                        memset(line,'q'+0x20,b); line[b]=0;
                        if (i == 0) {
                                line[0] = 'l'+0x20;
                                line[b-1] = 'k'+0x20;
                                memcpy(line+2,name,(int)strlen(name)+4>b ?
                                        b-4 : strlen(name));
                                *line = '*';
                                }
                        else {  line[0] = 'm'+0x20;
                                line[b-1] = 'j'+0x20;
                                }
                        }
                else {  memset(line,' ',b); line[b]=0;
                        line[0] = 'x'+0x20; line[b-1] = 'x'+0x20;
                        }
                if ((i >= ExY) && (i < ExY+ExH)) {
                        line[ExX+ExB] = 0;
                        tw_term_puts(View::RealX(ExX),View::RealY(i),TW_NORMAL,line+ExX);
                        }
                if (i >= ExY+ExH) break;
                }
	free(line);

	if (!ExX) ExX = 1;
	if (!ExY) ExY = 1;
        if (ExB+ExX > b-1) ExB = (b-1)-ExX;
        if (ExH+ExY > h-1) ExH = (h-1)-ExY;

	if (panner) {
		ex = xo+ExX-1;
		ey = yo+ExY-1;
		eb = ExB;
		eh = ExH;
                        
		if ((eb > 0) && (eh > 0))
			panner->expose(ex, ey,eb,eh);
		}
        }

void ViewScroll::event(int EvX,int EvY,int EvKey)
{       if (((EvY == 0) && (EvX == 0) && (EvKey == TERM_MAUS1)) ||
            (EvKey == TERM_ESC)) {
                VRoot.focus->focus_out(EvKey);
                VRoot.focus = &VRoot;
                unmap();
                VRoot.set_cursor(0,0);
                return;
                }

        if ((EvY == 0) && (EvKey == TERM_MAUS1)) {
                parent->up(this);
                if ((View::RealX(EvX) != MausX) || (View::RealY(EvY) != MausY))
                        move(x+(MausX-View::RealX(EvX)),y+(MausY-View::RealY(EvY)));
		focus_in(TERM_MAUS1);
                return;
                }
        if (EvKey == TERM_TAB) {
                VRoot.focus->focus_out(EvKey);
                VRoot.focus = &VRoot;
                VRoot.set_cursor(0,0);
                return;
                }
	if ((EvY == h-1) && (EvX == b-1)) {
                if ((View::RealX(EvX) != MausX) || (View::RealY(EvY) != MausY))
                        resize(b+(MausX-View::RealX(EvX)),h+(MausY-View::RealY(EvY)));
		return;
		}

	if (EvX && EvY && (EvX<b-1) && (EvY<h-1) && panner)
		panner->event(EvX+xo-1,EvY+yo-1,EvKey);
	}

void ViewScroll::resize(int SizB,int SizH)
{       int vis;
	View *par;

	vis = visible; par = parent;

	if (SizB && SizH) {
		if (vis) unmap();
                b = SizB;
                h = SizH;

                if (vis) map(par);
                }
        }

void ViewScroll::focus_in(int reason)
{	if (panner) {
		VRoot.focus= &VRoot;

		if ((xo != 1) || (yo != 1)) {
			yo=0;
			xo=0;
			VRoot.expose(View::RealX(0),View::RealY(0),b,h);
			}
		panner->focus_in(reason);
		}
	}

/* -------------------------------------------------------------------------- */

ViewFramedMask::ViewFramedMask(char *InitName, char **InitBackground,
		MaskLoad *InitLoad,
                int InitX, int InitY, int InitB, int InitH)
	: ViewScroll(InitName, 0, InitX,InitY,InitB,InitH)
{	panner = new ViewMask(InitName, InitBackground, InitLoad,
			InitX,InitY,InitB,InitH);

        if (panner) {
                panner->x = 0;
                panner->y = 0;
                panner->parent = this;
                if (!panner->b) panner->b = InitB-2;
                if (!panner->h) panner->h = InitH-2;
		if (!InitB) b = panner->b +2;
		if (!InitH) h = panner->h +2;
                }
	}

static ViewFramedMask *firstmask = 0;

void ViewFramedMask::map(View *par)
{	if (!firstmask) {
		firstmask = this;
		border = 0;
		header = 0;
		footer = 1;
		}
	ViewScroll::map(par);
	}

void ViewFramedMask::unmap()
{	if (firstmask == this) {
		firstmask = 0;
		border = 1;
		header = 0;
		footer = 0;

                if (panner) {
			panner->y = 0;
			b = 0;
			h = 0;
			}
		yo = 0; yo = 0;
		}
	ViewScroll::unmap();
	}

ViewFramedText::ViewFramedText(char *InitName, char **InitBackground,
                int InitX, int InitY, int InitB, int InitH)
	: ViewScroll(InitName, 0, InitX,InitY,InitB,InitH)
{	cur_x = 1;
	cur_y = 1;

	panner = new ViewBack(InitName, InitX,InitY,InitB,InitH);

        if (panner) {
        	((ViewBack*)(panner))->background = InitBackground;
                panner->x = 0;
                panner->y = 0;
                panner->parent = this;
                if (!panner->b) panner->b = InitB-2;
                if (!panner->h) panner->h = InitH-2;
		if (!InitB) b = panner->b +2;
		if (!InitH) h = panner->h +2;
                }
	}

void ViewFramedText::event(int EvX,int EvY,int EvKey)
{	set_cursor(cur_x,cur_y);

	if (EvKey == TERM_UP) {
		if (cur_y>0) cur_y --;
		set_cursor(cur_x,cur_y);
		}
	else
	if (EvKey == TERM_DN) {
		if (cur_y<panner->h) cur_y ++;
		set_cursor(cur_x,cur_y);
		}
	else
	ViewScroll::event(EvX,EvY,EvKey);
	}	

void ViewFramedText::focus_in(int reason)
{	if (panner) {
		VRoot.focus= &VRoot;
		set_cursor(cur_x,cur_y);
		panner->focus_in(reason);
		}
	}

/* -------------------------------------------------------------------------- */

ViewCursor::ViewCursor(char *InitName, char **InitBackground,
		MaskLoad *InitLoad, List *InitVcds,
                int InitX, int InitY, int InitB, int InitH)
	: ViewFramedMask(InitName, InitBackground, InitLoad, 
		InitX,InitY,InitB,InitH)
{	vcds = InitVcds;
	find_vcd(0);
	}

struct ViewCursor_def* ViewCursor::find_vcd(int line)
{	struct ViewCursor_def **v;
	int exp_flag =0;

	v = (struct ViewCursor_def**)vcds->lines;
	while (*v && v[1] && (v[1]->first<=line)) v++;

	/*
	fprintf(stderr,"\r\nfind_vcd %s\r\n",*v ? (*v)->cursor->name : "null");
	*/

	if (*v) {
		if ((vcd != *v) && visible) exp_flag++;

		vcd=*v;
		transaction=vcd->transaction;
		}
	else {	if ((vcd != *v) && visible) exp_flag++;

		vcd=0;
		transaction=0;
		}

	if (exp_flag) this->expose(0,h-1,b,1);

	return(*v);
	}

void ViewCursor::event(int EvX,int EvY,int EvKey)
{	int cur_y;

	cur_y = stdterm->y;

	find_vcd(EvY-1);

	if (vcd && (EvKey == TERM_PGUP)) {
		if (VRoot.focus != &VRoot) VRoot.focus->focus_out(EvKey);
		vcd->cursor->prev();
		if (vcd->check) (*(vcd->check))((ViewField *)this,EvKey);
		this->full_expose();
		}
	else
	if (vcd && (EvKey == TERM_PGDN)) {
		if (VRoot.focus != &VRoot) VRoot.focus->focus_out(EvKey);
		vcd->cursor->next();
		if (vcd->check) (*(vcd->check))((ViewField *)this,EvKey);
		this->full_expose();
		}
	else
	if (vcd && (EvKey >= TERM_F1) && (EvKey <= TERM_F10)) {
		if (VRoot.focus != &VRoot) VRoot.focus->focus_out(EvKey);
		if (vcd->check) (*(vcd->check))((ViewField *)this,EvKey);
		}
	else {
		ViewScroll::event(EvX,EvY,EvKey);
		if (cur_y != stdterm->y) {
			EvY = (stdterm->y-cur_y)+EvY;
			if ((EvY>=0) && (EvY<b))
			    	find_vcd(EvY-1);
			}
		}
	}

void ViewCursor::expose(int ExX,int ExY,int ExB,int ExH)
{	char line[80];

	ViewFramedMask::expose(ExX,ExY,ExB,ExH);
	if ((ExY+ExH >= h) && (ExX+ExB >= b) && vcd && vcd->cursor) {
		sprintf(line,"%s %d/%d",
			vcd->cursor->name,
			vcd->cursor->seek+1,
			vcd->cursor->nrecs);
		tw_term_puts(View::RealX(b-2-strlen(line)),View::RealY(h-1),TW_NORMAL,line);
		}
	}

void ViewCursor::focus_in(int reason)
{	ViewScroll::focus_in(reason);
	if (vcd && vcd->check) (*(vcd->check))((ViewField *)this,TERM_CR);
	}
